@@ -63,4 +63,10 @@ class TaskError(Exception):
     '''
     Raised when operation ordered in an improper task.
     '''
+    pass
+
+class LabelValueConflictError(Exception):
+    '''
+    Raised when a value represent more than one labels.
+    '''
     pass
\ No newline at end of file
