@@ -6,6 +6,7 @@ Created on Wed Aug 26 20:07:43 2020
 """
 from .exceptions import ModeError, LackOfParameterError, BrokenTimestepError,\
     KFoldError, LackOfLabelDictError, TaskError
+from .labeldict import BaseDict
 
 import math
 import json
@@ -14,6 +15,7 @@ import logging
 import numpy as np
 import tableprint as tp
 from sklearn.model_selection import ShuffleSplit
+from typing import Dict, Union
 
 def flatten(seq):
     return [i for a in seq for i in a]
@@ -22,7 +24,7 @@ class Sample(object):
     def __init__(self, unit='epoch', tmin=0, tmax=0, n_splits=10,
                  test_size=0.1, class_balance=True, data_balance=False,
                  epoch_padding=False, data_padding=None,
-                 task='classification'):
+                 task='classification', concat=False):
         self.__editable = True
         self.__autoencoder = False
         self.set_unit(unit)
@@ -35,6 +37,7 @@ class Sample(object):
         self.set_data_padding(data_padding)
         self.set_task(task)
         self.array_type = int
+        self.concat = concat
 
     def set_unit(self, unit):
         '''
@@ -173,7 +176,7 @@ class Sample(object):
         else:
             self.task = task.lower()
 
-    def set_label_array_type(self, array_type):
+    def set_label_array_type(self, array_type:type):
         '''
         Set the type of one-hot or N-hot label array the sample will return.
         If `array_type` is `None`, the function will return a number directly,
@@ -206,7 +209,31 @@ class Sample(object):
         '''
         self.array_type = array_type
 
-    def set_x(self, x):
+    def set_concat(self, concat:bool=True):
+        '''
+        Decide the output shape when tmin!=0 or tmax!=0, i.e., multiple time
+        steps are sampled.
+        If you have not set the `tmin` or `tmax`. You do not need to care about
+        this parameter.
+        The default is False, means the output will not concatenate time steps
+        (i.e. sampled epochs) but leave the output shape as
+                   `(time_step, channel[, feature-related])`
+        In this situation, every
+        time step occupies a unique matrix. It is suitable for RNN-related
+        models.
+        If `concat==True`, all the sampled epoches will concatenate by the
+        'time' axis. It results in continuous feature which is suitable for TCN
+        or other CNN-related models.
+
+        Parameters
+        ----------
+        concat : bool, optional
+            The default is True.
+
+        '''
+        self.concat = concat
+
+    def set_x(self, x, labeldict: Dict[str, BaseDict] = {}):
         # 不强求类型。可以是dataset的label甚至是condition(比如用很多condition来判断疾病(age, sex)→diagnose
         self.__check_editable()
         if isinstance(x, str):
@@ -215,6 +242,7 @@ class Sample(object):
             self.x = x
         if self.__autoencoder:
             self.set_y(self.get_x())
+        self.x_dict = labeldict
 
     def get_x(self):
         try:
@@ -222,7 +250,7 @@ class Sample(object):
         except:
             return None
             
-    def set_y(self, y):
+    def set_y(self, y, labeldict: Dict[str, BaseDict] = {}):
         # unit是data就不是epoched，unit是epoch就是epoched。或者说，如果unit是data，就从condition里找，如果unit是epoch就从label里找。
         self.__check_editable()
         if isinstance(y, str):
@@ -233,6 +261,7 @@ class Sample(object):
             self.get_x(): # assert x == y when using autoencoder
             raise ValueError('When using autoencoder, `y` must be `x` itself.')
         self.y = y
+        self.y_dict = labeldict
 
     def get_y(self):
         try:
@@ -335,20 +364,20 @@ class Sample(object):
         if not self.__preparation:
             self.init()
 
-    def __check_label_dict(self, dataset):
+    def __check_labeldict(self, dataset):
         if not self.__autoencoder:
             for x in self.get_x():
                 if dataset.elements[x] == 'label' and x not in \
-                    dataset.label_dict:
+                    dataset.labeldict:
                     raise LackOfLabelDictError('The element `{0}` does not '
-                                               'have a label_dict. Please use '
+                                               'have a labeldict. Please use '
                                                '`Dataset.one_hot()` function '
                                                'to create one or manually set '
                                                'one.'.format(x))
             for y in self.get_y():
-                if y not in dataset.label_dict:
+                if y not in dataset.labeldict:
                     raise LackOfLabelDictError('The element `{0}` does not '
-                                               'have a label_dict. Please use '
+                                               'have a labeldict. Please use '
                                                '`Dataset.one_hot()` function '
                                                'to create one or manually set '
                                                'one.'.format(y))
@@ -494,7 +523,7 @@ class Sample(object):
 
             # check dataset
             self.__check_dataset(dataset)
-            self.__check_label_dict(dataset)
+            self.__check_labeldict(dataset)
         except Exception as e:
             del self.mode
             raise e
@@ -502,35 +531,48 @@ class Sample(object):
         # set dataset
         self.dataset = dataset
         self.dataset.shape_check()
-        if data_selection is None and self.get_unit() == 'epoch':
-            self.__selection = None
-            self.data_selection = self.dataset.select_epochs()
-        elif data_selection is None and self.get_unit() == 'data':
-            self.__selection = None
-            self.data_selection = self.dataset.select_data()
+        # === SELECT DATA ===
+        # # use all the data or epochs
+        if data_selection is None:
+            if self.get_unit() == 'epoch':
+                self.__selection = None
+                self.data_selection = self.dataset.select_epochs()
+            elif self.get_unit() == 'data':
+                self.__selection = None
+                self.data_selection = self.dataset.select_data()
+        # # select data
         else:
             # data_selection has parameter in form of dict
             self.__selection = data_selection
             self.data_selection = set()
+            # case: epoch
             if self.get_unit() == 'epoch':
-                for e in data_selection.keys():
-                    if self.data_selection and len(data_selection[e]) != 0:
+                for e in data_selection.keys():  # for each element
+                    # [1] there is already a subset of epochs in the
+                    # `self.data_selection` and the element `e` provides some
+                    # selection range.
+                    if self.data_selection and len(self.__selection[e]) != 0:
                         self.data_selection.intersection_update(
-                            self.dataset.select_epochs(e, data_selection[e]))
+                            set(self.dataset.select_epochs(e, data_selection[e]))
+                        )
+                    # [2] `self.data_selection` is an empty set
                     elif (not self.data_selection) and \
                         len(data_selection[e]) != 0:
                         self.data_selection = \
-                            self.dataset.select_epochs(e, data_selection[e])
+                            set(self.dataset.select_epochs(e, data_selection[e]))
+                    # [3] other situation
                     elif self.data_selection:
                         pass
                     else:
                         self.data_selection = \
                             self.dataset.select_epochs()
-            else:
-                for e in data_selection.keys():
+            # case: data
+            elif self.get_unit() == 'data':
+                for e in data_selection.keys():  # for each element
                     if self.data_selection and len(data_selection[e]) != 0:
                         self.data_selection.intersection_update(
-                            self.dataset.select_data(e, data_selection[e]))
+                            set(self.dataset.select_data(e, data_selection[e]))
+                        )
                     elif (not self.data_selection) and \
                         len(data_selection[e]) != 0:
                         self.data_selection = \
@@ -576,6 +618,16 @@ class Sample(object):
         Initialize all the steps before sampling, including padding and
         spliting the train_set and test_set.
         '''
+        logging.debug('== CLEANING VARIABLES ==')
+        if hasattr(self, 'train'):
+            del self.train
+        if hasattr(self, 'test'):
+            del self.test
+        if hasattr(self, 'onesamp'):
+            del self.onesamp
+        if hasattr(self, 'opc'):  # one per class
+            del self.opc
+
         logging.info('== INITIALIZING SAMPLING ==')
         if self.task == 'classification':
             self.subgroups()
@@ -845,7 +897,10 @@ class Sample(object):
                         tmax=self.get_tmax(), 
                         epoch_padding=self.epoch_padding,
                         autoencoder=self.__autoencoder,
-                        array_type=self.array_type)
+                        array_type=self.array_type,
+                        concat=self.concat,
+                        x_dict = self.x_dict,
+                        y_dict = self.y_dict)
                 elif self.get_unit() == 'data' and not self.disturb:
                     yield self.dataset.sample_data(
                         item,
@@ -856,7 +911,10 @@ class Sample(object):
                         max_len=self.max_len,
                         epoch_padding=self.epoch_padding,
                         autoencoder=self.__autoencoder,
-                        array_type=self.array_type)
+                        array_type=self.array_type,
+                        concat=self.concat,
+                        x_dict = self.x_dict,
+                        y_dict = self.y_dict)
                 elif self.get_unit() == 'epoch' and self.disturb:
                     yield self.dataset.sample_epoch(
                         item[0],
@@ -868,7 +926,10 @@ class Sample(object):
                         test_data_name=self.train_y[idx][0],
                         test_epoch=self.train_y[idx][1],
                         autoencoder=self.__autoencoder,
-                        array_type=self.array_type)
+                        array_type=self.array_type,
+                        concat=self.concat,
+                        x_dict = self.x_dict,
+                        y_dict = self.y_dict)
                 elif self.get_unit() == 'data' and not self.disturb:
                     yield self.dataset.sample_data(
                         item,
@@ -880,7 +941,10 @@ class Sample(object):
                         epoch_padding=self.epoch_padding,
                         test_data_name=self.train_y[idx],
                         autoencoder=self.__autoencoder,
-                        array_type=self.array_type)
+                        array_type=self.array_type,
+                        concat=self.concat,
+                        x_dict = self.x_dict,
+                        y_dict = self.y_dict)
             except BrokenTimestepError:
                 continue
 
@@ -913,7 +977,10 @@ class Sample(object):
                         tmax=self.get_tmax(), 
                         epoch_padding=self.epoch_padding,
                         autoencoder=self.__autoencoder,
-                        array_type=self.array_type)
+                        array_type=self.array_type,
+                        concat=self.concat,
+                        x_dict = self.x_dict,
+                        y_dict = self.y_dict)
                 elif self.get_unit() == 'data' and not self.disturb:
                     yield self.dataset.sample_data(
                         item,
@@ -924,7 +991,10 @@ class Sample(object):
                         max_len=self.max_len,
                         epoch_padding=self.epoch_padding,
                         autoencoder=self.__autoencoder,
-                        array_type=self.array_type)
+                        array_type=self.array_type,
+                        concat=self.concat,
+                        x_dict = self.x_dict,
+                        y_dict = self.y_dict)
                 elif self.get_unit() == 'epoch' and self.disturb:
                     yield self.dataset.sample_epoch(
                         item[0],
@@ -936,7 +1006,10 @@ class Sample(object):
                         test_data_name=self.test_y[idx][0],
                         test_epoch=self.test_y[idx][1],
                         autoencoder=self.__autoencoder,
-                        array_type=self.array_type)
+                        array_type=self.array_type,
+                        concat=self.concat,
+                        x_dict = self.x_dict,
+                        y_dict = self.y_dict)
                 elif self.get_unit() == 'data' and self.disturb:
                     yield self.dataset.sample_data(
                         item,
@@ -948,7 +1021,10 @@ class Sample(object):
                         epoch_padding=self.epoch_padding,
                         test_data_name=self.test_y[idx],
                         autoencoder=self.__autoencoder,
-                        array_type=self.array_type)
+                        array_type=self.array_type,
+                        concat=self.concat,
+                        x_dict = self.x_dict,
+                        y_dict = self.y_dict)
             except BrokenTimestepError:
                 continue
 
@@ -1002,9 +1078,10 @@ class Sample(object):
         # the purpose of `Sample.one` is only to get one sample, regardless of
         # whether the mode `train` or `test`, or whether the dataset needs to
         # cross validation or not.
-        one = random.sample(self.data_selection, 1)
-        one = one * len(self.data_selection)
-        for idx, item in enumerate(one):
+        if not hasattr(self, 'onesamp'):
+            onesamp = random.sample(self.data_selection, 1)
+            self.onesamp = onesamp * len(self.data_selection)
+        for idx, item in enumerate(self.onesamp):
             try:
                 if self.get_unit() == 'epoch':
                     yield self.dataset.sample_epoch(
@@ -1015,7 +1092,10 @@ class Sample(object):
                         tmax=self.get_tmax(), 
                         epoch_padding=self.epoch_padding,
                         autoencoder=self.__autoencoder,
-                        array_type=self.array_type)
+                        array_type=self.array_type,
+                        concat=self.concat,
+                        x_dict = self.x_dict,
+                        y_dict = self.y_dict)
                 elif self.get_unit() == 'data':
                     yield self.dataset.sample_data(
                         item,
@@ -1026,7 +1106,10 @@ class Sample(object):
                         max_len=self.max_len,
                         epoch_padding=self.epoch_padding,
                         autoencoder=self.__autoencoder,
-                        array_type=self.array_type)
+                        array_type=self.array_type,
+                        concat=self.concat,
+                        x_dict = self.x_dict,
+                        y_dict = self.y_dict)
             except BrokenTimestepError:
                 continue
     
@@ -1085,17 +1168,18 @@ class Sample(object):
         np.ndarray
 
         '''
-        if not hasattr(self, 'classes'):
-            self.subgroups()
-        # the purpose of `Sample.one_per_class` is only to get samples,
-        # regardless of whether the mode `train` or `test`, or whether the
-        # dataset needs to cross validation or not.
-        opc = [] # one_per_class
-        for k in self.classes.keys():
-            opc.extend(random.sample(self.classes[k], 1))
-        # the length of samples = data_length
-        opc = opc * math.floor(
-            len(self.data_selection) / len(self.classes.keys()))
+        if not hasattr(self, 'opc'):
+            if not hasattr(self, 'classes'):
+                self.subgroups()
+            # the purpose of `Sample.one_per_class` is only to get samples,
+            # regardless of whether the mode `train` or `test`, or whether the
+            # dataset needs to cross validation or not.
+            opc = [] # one_per_class
+            for k in self.classes.keys():
+                opc.extend(random.sample(self.classes[k], 1))
+            # the length of samples = data_length
+            self.opc = opc * math.floor(
+                len(self.data_selection) / len(self.classes.keys()))
         for idx, item in enumerate(opc):
             try:
                 if self.get_unit() == 'epoch':
@@ -1107,7 +1191,10 @@ class Sample(object):
                         tmax=self.get_tmax(), 
                         epoch_padding=self.epoch_padding,
                         autoencoder=self.__autoencoder,
-                        array_type=self.array_type)
+                        array_type=self.array_type,
+                        concat=self.concat,
+                        x_dict = self.x_dict,
+                        y_dict = self.y_dict)
                 elif self.get_unit() == 'data':
                     yield self.dataset.sample_data(
                         item,
@@ -1118,7 +1205,10 @@ class Sample(object):
                         max_len=self.max_len,
                         epoch_padding=self.epoch_padding,
                         autoencoder=self.__autoencoder,
-                        array_type=self.array_type)
+                        array_type=self.array_type,
+                        concat=self.concat,
+                        x_dict = self.x_dict,
+                        y_dict = self.y_dict)
             except BrokenTimestepError:
                 continue
 
